/**
  *  \file client/cm-standard.cc
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>                // sin, cos
#include "client/cm-standard.h"
#include "client/chartloc.h"
#include "client/screens.h"
#include "client/tiles.h"
#include "cpluslib/format.h"
#include "game/formula.h"       // PI
#include "game/global.h"
#include "game/planet.h"
#include "game/ship.h"
#include "gfx/gfxcompl.h"
#include "gfx/palette.h"
#include "gfx/update.h"
#include "int/if/planetif.h"
#include "int/if/shipif.h"
#include "int/if/userif.h"
#include "int/intglobal.h"
#include "int/keymap.h"
#include "int/simple.h"
#include "ui/event.h"
#include "ui/ui.h"
#include "util/nls.h"
#include "ui/uiglobal.h"
#include "client/marks.h"
#include "game/lock.h"
#include "client/dialogs/labeldlg.h"
#include "client/dialogs/gotoxydlg.h"
#include "client/dlg-draw.h"
#include "client/cm-marker.h"
#include "client/cm-distance.h"
#include "client/cm-message.h"
#include "client/screens/ionscreen.h"
#include "client/cm-select.h"
#include "client/widgets/markertype.h"
#include "client/widgets/colorselect.h"
#include "client/dialogs/chartconfig.h"
#include "client/markerconfig.h"

static void drawFancyBar(GfxCanvas& can, GfxRect r);

class WStarchartTileSkin : public UISkin {
 public:
    WStarchartTileSkin()
        : UISkin(DARK_COLOR_SET)
        { }
    void  drawBackground(GfxContext& ctx, GfxRect r);
};

void
WStarchartTileSkin::drawBackground(GfxContext& /*ctx*/, GfxRect /*r*/)
{
    // FIXME: empty?
}

class WStarchartTileContainer : public UIBaseWidget, public WTileContainer {
    ObjectHolder h;
    WStarchartTileSkin skin;
 public:
    WStarchartTileContainer();

    virtual void          drawContent(GfxCanvas& canvas);                       // draw content of widget, end-users override this
    virtual void          drawEverything(GfxCanvas& canvas);                    // draw everything including children, bases override this
    virtual void          drawBelow(UIBaseWidget& child, GfxCanvas& canvas);    // draw everything below child, bases override this
    virtual GfxCanvas&    getCanvasFor(UIBaseWidget& child, ObjectHolder& mgr); // get canvas to draw child on, bases override this
    virtual bool          handleEvent(const UIEvent& event, bool second_pass);   // can be overridden by whoever needs it, base provides default
    virtual void          onAddRemoveChild(UIBaseWidget& the_widget, bool added); // overridden in base
    virtual void          getLayoutInfo(LayoutInfo&);                          // overridden by everyone
    virtual void          onResize(GfxRect oldsize, bool by_layout);           // overridden in base
    virtual void          onChildResize(UIBaseWidget& child, GfxRect oldsize); // overridden in base
    virtual int           run();                 // overridden in root
    virtual void          stop(int arg);         // overridden in root

    void doLayout(GChartViewport& viewport);
    void addNewTile(WTile* w);
};

WStarchartTileContainer::WStarchartTileContainer()
{
    setSkin(skin);
}

void
WStarchartTileContainer::drawContent(GfxCanvas& canvas)
{
    drawFancyBar(canvas, getExtent());
    canvas.update(getExtent());
}

void
WStarchartTileContainer::drawEverything(GfxCanvas& canvas)
{
    drawContent(canvas);
    for (UIBaseWidget* p = getFirstChild(); p != 0; p = p->getNextSibling())
        p->drawEverything(canvas);
}

void
WStarchartTileContainer::drawBelow(UIBaseWidget& /*child*/, GfxCanvas& /*canvas*/)
{
    // draw everything below child, bases override this
}

GfxCanvas&
WStarchartTileContainer::getCanvasFor(UIBaseWidget& /*child*/, ObjectHolder& /*mgr*/)
{
    // get canvas to draw child on, bases override this
    return *ui_screen;
}

bool
WStarchartTileContainer::handleEvent(const UIEvent& /*event*/, bool /*second_pass*/)
{
    // can be overridden by whoever needs it, base provides default
    return false;
}

void
WStarchartTileContainer::onAddRemoveChild(UIBaseWidget& /*the_widget*/, bool /*added*/)
{
    /* FIXME: at first, we assume no uncontrolled add/remove.
       Otherwise, we'd simply have to request relayout.  */
}

void
WStarchartTileContainer::getLayoutInfo(LayoutInfo& li)
{
    li.min_size = li.pref_size = GfxPoint(0, 0);
    li.flags = li_NoLayout;
}

void
WStarchartTileContainer::onResize(GfxRect /*oldsize*/, bool /*by_layout*/)
{
    // Nothing to do?
}

void
WStarchartTileContainer::onChildResize(UIBaseWidget& /*child*/, GfxRect /*oldsize*/)
{
    /* FIXME: at first, we assume no uncontrolled add/remove/resize.
       Otherwise, we'd simply have to request relayout.  */
}

int
WStarchartTileContainer::run()
{
    /* does not run itself */
    return 0;
}

void
WStarchartTileContainer::stop(int /*arg*/)
{ /* does not run itself */ }

void
WStarchartTileContainer::doLayout(GChartViewport& viewport)
{
    enum { EDGE = 5 };

    // Compute total height and width
    int total_height = EDGE;
    int min_width    = 0;
    for (UIBaseWidget* w = getFirstChild(); w != 0; w = w->getNextSibling()) {
        LayoutInfo li;
        w->getLayoutInfo(li);
        total_height += EDGE + li.pref_size.y;
        if (li.pref_size.x > min_width)
            min_width = li.pref_size.x;
    }
    min_width += 2*EDGE;

    // Set own size and allocate space
    GfxRect myExtent(0, 0, min_width, total_height);
    myExtent.moveToEdge(viewport.view, 2, 1, 10);
    setExtent(myExtent, true);

    // Move children
    int y = myExtent.y + EDGE;
    for (UIBaseWidget* w = getFirstChild(); w != 0; w = w->getNextSibling()) {
        LayoutInfo li;
        w->getLayoutInfo(li);
        w->setExtent(GfxRect(myExtent.x + EDGE, y, myExtent.w - 2*EDGE, li.pref_size.y), false);
        y += li.pref_size.y + EDGE;
    }
}

void
WStarchartTileContainer::addNewTile(WTile* w)
{
    addChildWidget(h.add(w), getLastChild());
}

static void
drawFancyBar(GfxCanvas& can, GfxRect r)
{
    const int icolor = standard_colors[COLOR_SHIELD+3];
    const int fcolor = standard_colors[COLOR_SHIELD+5];
    const int z = standard_colors[COLOR_BLACK];

    int alpha;

    if (ui_screen_bits <= 8) {
        can.drawBar(r, icolor, z, GRAY50_FILL, GFX_ALPHA_OPAQUE);
        alpha = GFX_ALPHA_OPAQUE;
    } else {
        can.drawBar(r, icolor, z, SOLID_FILL, 128);
        alpha = (3*GFX_ALPHA_OPAQUE + GFX_ALPHA_TRANSPARENT)/4;
    }

    can.drawBar(GfxRect(r.x-1,   r.y-2,   r.w+2, 2),   fcolor, z, SOLID_FILL, alpha);
    can.drawBar(GfxRect(r.x-1,   r.y+r.h, r.w+2, 2),   fcolor, z, SOLID_FILL, alpha);
    can.drawBar(GfxRect(r.x-2,   r.y,     2,     r.h), fcolor, z, SOLID_FILL, alpha);
    can.drawBar(GfxRect(r.x+r.w, r.y,     2,     r.h), fcolor, z, SOLID_FILL, alpha);
    can.drawPixel(GfxPoint(r.x-2,     r.y-1),   fcolor, alpha);
    can.drawPixel(GfxPoint(r.x+r.w+1, r.y-1),   fcolor, alpha);
    can.drawPixel(GfxPoint(r.x-2,     r.y+r.h), fcolor, alpha);
    can.drawPixel(GfxPoint(r.x+r.w+1, r.y+r.h), fcolor, alpha);
}

static void
drawCursor(GfxCanvas& can, GfxPoint pt, int angle)
{
    static const int SIZE = 15;
    static const double FACTOR = 5 * PI / 180;

    angle = -angle;

    GfxContext ctx;
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.useCanvas(can);
    ctx.setColor(COLOR_RED);

    /* it turns out that scaling the cursor doesn't look that good. */
    const double scaled_size = SIZE;

    for (int i = 0; i < 2; ++i) {
        double d = angle * PI / 180;
        drawLine(ctx,
                 GfxPoint(pt.x + roundToInt(4*std::sin(d)), pt.y + roundToInt(4*std::cos(d))),
                 GfxPoint(pt.x + roundToInt(scaled_size*std::sin(d)), pt.y + roundToInt(scaled_size*std::cos(d))));
        for (int j = -8; j < 8; ++j) {
            drawLine(ctx,
                     GfxPoint(pt.x + roundToInt(scaled_size*std::sin(d-FACTOR*j)),
                              pt.y + roundToInt(scaled_size*std::cos(d-FACTOR*j))),
                     GfxPoint(pt.x + roundToInt(scaled_size*std::sin(d-FACTOR*j+FACTOR)),
                              pt.y + roundToInt(scaled_size*std::cos(d-FACTOR*j+FACTOR))));
        }
        angle += 180;
    }
}

static void
drawObjectList(GfxCanvas& can, GChartLocation& loc)
{
    const GObjectList& chart_objlist = loc.getObjectList();

    if (!chart_objlist.getNumObjects())
        return;

    int lines = loc.getObjectListHeight();
    int top   = loc.getObjectListTop();

    GfxPoint pt = loc.getViewport().getCenter();
    int h = font_heights[FONT_NORMAL]*lines;
    int y = pt.y - h/2;

    drawFancyBar(can, GfxRect(pt.x - 300, y-2, 280, h+4));

    GfxContext ctx;
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);
    for (int i = 0; i < lines; ++i) {
        GObjectId id = chart_objlist.getObjectId(i + top);
        UIColor::Color color = (id.isGameObject()
                                ? getPlayerColor(id.getObject().getOwner())
                                : UIColor::tc_Static);
        if (id == loc.getCurrentObject()) {
            if (ui_screen_bits <= 8)
                can.drawBar(GfxRect(pt.x-298, y+i*font_heights[FONT_NORMAL], 276, font_heights[FONT_NORMAL]),
                            standard_colors[COLOR_SHIELD+4], GFX_TRANSPARENT, SOLID_FILL, GFX_ALPHA_OPAQUE);
            else
                can.drawBar(GfxRect(pt.x-298, y+i*font_heights[FONT_NORMAL], 276, font_heights[FONT_NORMAL]),
                            standard_colors[COLOR_SHIELD+8], 0, SOLID_FILL, 128);
        }

        int wi = 276;
        if (id.isGameObject() && id.getObject().isMarked()) {
            ctx.setColor(COLOR_YELLOW);
            drawSelection(ctx, GfxPoint(pt.x-wi-17, y+i*font_heights[FONT_NORMAL]+7), 1, 2);
            wi -= 15;
        }
        ctx.setColor(BLACK_COLOR_SET[color]);
        outTextF(ctx, pt.x-wi-22, y+i*font_heights[FONT_NORMAL], wi, id.getName());
    }
}

/** Page through objects.
    \param loc Chart location object to operate on
    \param delta Browse direction (+1/-1)
    \param marked_only true to accept only marked objects
    \param any true to accept any race's object. Default is to accept only played objects
           when starting at a played object.

    FIXME: PCC 1.x behaves a little different. With any=false, PCC 1.x would only accept
    played objects no matter where we start, but it would "degrade" to accept non-played
    objects when there are no played ones. */
static void
doPage(GChartLocation& loc, int delta, bool marked_only, bool any)
{
    GObjectType* type;

    /* Fetch current object, and associated type object */
    GObjectId id = loc.getCurrentObject();
    if (!id.isValid() || !id.isGameObject())
        return;

    if (GShip* sh = dynamic_cast<GShip*>(&id.getObject())) {
        // Iterate through all ships or player ships
        if (sh->isPlayed() && !any)
            type = &global_ship_type;
        else
            type = &global_anyship_type;
    } else if (GPlanet* pl = dynamic_cast<GPlanet*>(&id.getObject())) {
        // Iterate through all planets or player planets
        if (pl->isPlayed() && !any)
            type = &global_planet_type;
        else
            type = &global_anyplanet_type;
    } else {
        // Iterate through whatever it is
        type = &id.getType();
    }

    /* Get next object */
    int next;
    if (delta > 0)
        next = type->getNextWrap(id.getIndex(), marked_only);
    else
        next = type->getPreviousWrap(id.getIndex(), marked_only);

    /* Valid? */
    if (type->isExistant(next))
        loc.setCurrentObject(type->getObject(next));
}

static void
doChartLock(GChartLocation& loc, TLockMode lm, bool marked)
{
    GPoint pt = loc.getLocation();
    GLockData locker(pt, marked);
    findItem(getDisplayedTurn().getCurrentUniverse(), locker, lm);
    if (const GMapObject* obj = locker.getObject()) {
        loc.setCurrentObject(*obj);
    }
    loc.setLocation(locker.getPoint().getSimpleNearestAlias(loc.getLocation()));
}

static void
doGoTo(GChartLocation& loc)
{
    GPoint pt;
    if (doGoToXYDialog(pt))
        loc.setLocation(pt);
}

static void
doColorizeAdjacent(GDrawingContainer& c, GPoint pt, uint8 color)
{
    GDrawingContainer::iterator i = c.begin(), e = c.end();
    pt = pt.getSimpleCanonicalLocation();
    while (i != e) {
        if (i->isVisible() && i->getType() == LineDrawing && i->getColor() != color) {
            if (i->getPos().getSimpleCanonicalLocation() == pt) {
                // Starts at given position
                i->setColor(color);
                pt = i->getPos2().getSimpleCanonicalLocation();
                i = c.begin();
            } else if (i->getPos2().getSimpleCanonicalLocation() == pt) {
                // Ends at given position
                i->setColor(color);
                pt = i->getPos().getSimpleCanonicalLocation();
                i = c.begin();
            } else {
                // No match
                ++i;
            }
        } else {
            ++i;
        }
    }
}

static void
doColorChange(GChartLocation& loc)
{
    GChartViewport& viewport = loc.getViewport();

    // Not possible if drawings not visible
    if ((getChartOpts(false, viewport.mult, viewport.divi).show & GChartOptions::co_Drawings) == 0)
        return;

    // Find nearest visible drawing
    GDrawingContainer& c = getDisplayedTurn().getCurrentUniverse().getDrawings();
    GDrawingContainer::iterator i = c.findNearestVisibleDrawing(loc.getLocation().getSimpleCanonicalLocation());
    if (i == c.end())
        return;

    // Change color
    WColorSelector csel;
    csel.setColor(GDrawing::getUserColorFromColor(i->getColor()));
    int n = csel.doStandardDialog(_("Drawing Color"), (i->getType() == LineDrawing));
    if (n == csel.rCancel)
        return;

    uint8 new_color = GDrawing::getColorFromUserColor(csel.getColor());
    i->setColor(new_color);
    if (n == csel.rAdjacent) {
        doColorizeAdjacent(c, i->getPos(), new_color);
        doColorizeAdjacent(c, i->getPos2(), new_color);
    }

    loc.markChanged();
}

static void
doMoveMarker(GChartLocation& loc, WChartModeList& modes)
{
    // Not possible if drawings not visible
    GChartViewport& viewport = loc.getViewport();
    if ((getChartOpts(false, viewport.mult, viewport.divi).show & GChartOptions::co_Drawings) == 0)
        return;

    // Find drawing here
    GDrawingContainer& c = getDisplayedTurn().getCurrentUniverse().getDrawings();
    GDrawingContainer::iterator i = c.begin();
    while (i != c.end() && (i->getPos().getSimpleCanonicalLocation() != loc.getLocation().getSimpleCanonicalLocation() || i->getType() != MarkerDrawing))
        ++i;

    if (i != c.end())
        WMarkerChartMode::moveMarker(modes, i);
}

static void
doDraw(GChartLocation& loc, WChartModeList& modes)
{
    // Ask user what to do
    WNewDrawingDialog ndd;
    ndd.setColor(9);
    if (!ndd.runDialog())
        return;

    // Draw it
    GDrawing d(loc.getLocation().getSimpleCanonicalLocation(), ndd.getType());
    d.setColor(GDrawing::getColorFromUserColor(ndd.getColor()));
    d.setTag(ndd.getTag());
    switch (ndd.getType()) {
     case LineDrawing:
     case RectangleDrawing:
        d.setPos2(d.getPos());
        WMarkerChartMode::drawLineRectangle(modes, getDisplayedTurn().getCurrentUniverse().getDrawings().add(d));
        break;
     case MarkerDrawing: {
        WMarkerTypeSelector mts;
        WColorSelector cs;
        if (mts.doStandardDialog(_("New marker"))) {
            d.setMarkerKind(mts.getMarkerType());
            cs.setColor(GDrawing::getUserColorFromColor(d.getColor()));
            if (cs.doStandardDialog(_("New marker"), false) == cs.rOK) {
                d.setColor(GDrawing::getColorFromUserColor(cs.getColor()));
                getDisplayedTurn().getCurrentUniverse().getDrawings().add(d);
                loc.markChanged();
            }
        }
        delete modes.getModeByKind(ChartMode_PrimaryMode);
        break;
     }
     case CircleDrawing:
        d.setCircleRadius(10);
        WMarkerChartMode::drawCircle(modes, getDisplayedTurn().getCurrentUniverse().getDrawings().add(d));
        break;
    }
}

static void
doDistance(GChartLocation& loc, WChartModeList& modes)
{
    modes.pushMode(*new WDistanceChartMode(loc.getLocation(), loc.getCurrentShipId()));
    if (loc.getCurrentShipId() != 0) {
        GShip& sh = getDisplayedTurn().getCurrentUniverse().getShip(loc.getCurrentShipId());
        if (sh.getKind() == GShip::CurrentShip && sh.getPos() == loc.getLocation()) {
            loc.setLocation(sh.getWaypoint());
        }
    }
}

static void
doMarkRange(GChartLocation& loc, WChartModeList& modes)
{
    modes.pushMode(*new WSelectChartMode(loc.getLocation()));
}

static void
doDelete(GChartLocation& loc, WChartModeList& modes)
{
    GChartViewport& viewport = loc.getViewport();

    // Not possible if drawings not visible
    if ((getChartOpts(false, viewport.mult, viewport.divi).show & GChartOptions::co_Drawings) == 0)
        return;

    // Find nearest visible drawing
    GDrawingContainer& c = getDisplayedTurn().getCurrentUniverse().getDrawings();
    GDrawingContainer::iterator i = c.findNearestVisibleDrawing(loc.getLocation().getSimpleCanonicalLocation());
    if (i == c.end())
        return;

    WMarkerChartMode::deleteDrawing(modes, i);
}

static void
tryEditMarkerComment(GChartLocation& loc)
{
    /* Find a marker */
    GPoint pt = loc.getLocation().getSimpleCanonicalLocation();
    GDrawingContainer& drawings = getDisplayedTurn().getCurrentUniverse().getDrawings();
    GDrawingContainer::iterator p = drawings.begin();
    while (p != drawings.end()
           && !(p->isVisible()
                && p->getPos().getSimpleCanonicalLocation() == pt
                && p->getType() == MarkerDrawing))
        ++p;

    /* Edit it */
    if (p != drawings.end()) {
        editMarkerComment(*p);
        loc.markChanged();
    }
}

static void
createCannedMarker(int id, GChartLocation& loc)
{
    ConfigMarkerOption& opt = ConfigMarkerOption::getCannedMarker(id % ConfigMarkerOption::NUM_CANNED_MARKERS);

    // Create marker
    GDrawing marker(loc.getLocation().getSimpleCanonicalLocation(), MarkerDrawing);
    marker.setMarkerKind(opt.getType());
    marker.setColor(marker.getColorFromUserColor(opt.getColorIndex()));
    getDisplayedTurn().getCurrentUniverse().getDrawings().add(marker);

    // Redraw
    loc.markChanged();
}

/*************************** WStandardChartMode **************************/

/** Constructor. */
WStandardChartMode::WStandardChartMode()
    : WChartMode(ChartMode_BaseMode),
      cursor_angle(0),
      overlays(0)
{ }

/** Destructor. */
WStandardChartMode::~WStandardChartMode()
{
    delete overlays;
}

void
WStandardChartMode::onMove(GChartLocation& loc)
{
    // Build overlay tile
    delete overlays;
    overlays = 0;
    if (loc.getViewName().size() != 0) {
        overlays = new WStarchartTileContainer();
        createTileLayout(*overlays, loc.getViewName(), const_cast<GObjectList&>(loc.getObjectList()));
        overlays->doLayout(loc.getViewport());
    }

    // Do we have an overlay?
    if (overlays != 0) {
        if (overlays->getFirstChild() == 0) {
            delete overlays;
            overlays = 0;
        }
    }

    // Call next mode
    if (WChartMode* cm = getNextMode())
        cm->onMove(loc);
}


void
WStandardChartMode::drawBelow(GfxCanvas& can, GChartLocation& loc)
{
    // Call next mode
    if (WChartMode* cm = getNextMode())
        cm->drawBelow(can, loc);
}

void
WStandardChartMode::drawCursor(GfxCanvas& can, GChartLocation& loc)
{
    GfxPoint sc = loc.getViewport().scaleXY(loc.getLocation().getSimpleNearestAlias(loc.getCenter()));
    if (!loc.getCurrentObject().isValid()) {
        GfxContext ctx;
        ctx.useColorScheme(GfxStandardColorScheme::instance);
        ctx.useCanvas(can);
        ctx.setColor(COLOR_BLUE);
        drawHLine(ctx, sc.x-30, sc.y, sc.x-6);
        drawHLine(ctx, sc.x+30, sc.y, sc.x+6);
        drawVLine(ctx, sc.x, sc.y-30, sc.y-6);
        drawVLine(ctx, sc.x, sc.y+30, sc.y+6);
    } else {
        ::drawCursor(can, sc, cursor_angle);
    }
}

void
WStandardChartMode::drawOverlays(GfxCanvas& can, GChartLocation& loc)
{
    drawObjectList(can, loc);

    if (overlays) {
        GfxUpdateOptimizer o(can);
        GfxRect r;
        overlays->drawEverything(o);
        o.clearUpdateRegion(r); /* we forget the region, because the whole screen is refreshed afterwards */
    }

    // Coordinates
    GPoint pt = loc.getLocation().getSimpleCanonicalLocation();
    GPoint pt1 = pt.getCanonicalLocation();
    GfxRect view(loc.getViewport().view);

    GfxContext ctx;
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_FIXED]);
    ctx.setColor(COLOR_WHITE);
    ctx.setTextAlign(2, 2);
    if (pt == pt1) {
        outText(ctx, view.getRightX(), view.getBottomY(), format("%4d,%4d", pt.x, pt.y));
    } else {
        outText(ctx, view.getRightX(), view.getBottomY() - font_heights[FONT_NORMAL], format("%4d,%4d", pt.x, pt.y));
        outText(ctx, view.getRightX(), view.getBottomY(), format(_("wraps to %4d,%4d"), pt1.x, pt1.y));
    }

    // Sector Number
    int secn = pt.getSectorNumber();
    if (secn != 0) {
        ctx.setTextAlign(0, 2);
        outText(ctx, view.x, view.getBottomY(), format(_("Sector %d"), secn));
    }
}

bool
WStandardChartMode::handleEvent(const UIEvent& event, GChartLocation& loc)
{
    if (event.type == ev_Tick) {
        if (loc.getCurrentObject().isValid()) {
            /* Advance rotation, and trigger redraw. This used to draw
               in-place; the new way supports real interaction with
               other overlays. */
            cursor_angle = (cursor_angle + 3) % 360;
            GfxPoint sc = loc.getViewport().scaleXY(loc.getLocation());
            getParent()->markDirty(GfxRect(sc.x - 15, sc.y - 15, 31, 31));
        }
    } else if (event.type == ev_Key) {
        // Check keymap
        IntKeymapRef km = IntUserInterfaceBinding::getCurrentInstance()->getKeymap();

        // Do we have a key?
        uint32_t cmd;
        if (km && isAtom(cmd = km->lookupCommand(event.key.code))) {
            IntTemporaryContextLock lock;
            if (loc.getCurrentObject().isValid()) {
                if (dynamic_cast<GShip*>(&loc.getCurrentObject().getObject())) {
                    global_contexts.push_back(new IntShipContext(loc.getCurrentObject().getId()));
                } else if (dynamic_cast<GPlanet*>(&loc.getCurrentObject().getObject())) {
                    global_contexts.push_back(new IntPlanetContext(loc.getCurrentObject().getId()));
                } else {
                    // No object
                }
            }
            executeStatement(atomStr(cmd), event.key.arg);
            return true;
        } else {
            switch (event.key.code) {
             case SDLK_RETURN:
                doChartLock(loc, lm_Right, false);
                return true;
             case SDLK_SPACE:
                doChartLock(loc, lm_Left, false);
                return true;
             case ss_Ctrl + SDLK_RETURN:
                doChartLock(loc, lm_Right, true);
                return true;
             case ss_Ctrl + SDLK_SPACE:
                doChartLock(loc, lm_Left, true);
                return true;
             case '-':
             case vk_WheelDown:
                loc.doZoom(false);
                return true;
             case '+':
             case vk_WheelUp:
                loc.doZoom(true);
                return true;
             case SDLK_TAB:
                loc.doScroll(+1, false);
                return true;
             case ss_Shift + SDLK_TAB:
                loc.doScroll(-1, false);
                return true;
             case ss_Ctrl + SDLK_TAB:
                loc.doScroll(+1, true);
                return true;
             case ss_Shift + ss_Ctrl + SDLK_TAB:
                loc.doScroll(-1, true);
                return true;
             case SDLK_PAGEDOWN:
                doPage(loc, +1, false, false);
                return true;
             case SDLK_PAGEDOWN + ss_Ctrl:
                doPage(loc, +1, true, false);
                return true;
             case SDLK_PAGEUP:
                doPage(loc, -1, false, false);
                return true;
             case SDLK_PAGEUP + ss_Ctrl:
                doPage(loc, -1, true, false);
                return true;
             case SDLK_PAGEDOWN + ss_Shift:
                doPage(loc, +1, false, true);
                return true;
             case SDLK_PAGEDOWN + ss_Ctrl + ss_Shift:
                doPage(loc, +1, true, true);
                return true;
             case SDLK_PAGEUP + ss_Shift:
                doPage(loc, -1, false, true);
                return true;
             case SDLK_PAGEUP + ss_Ctrl + ss_Shift:
                doPage(loc, -1, true, true);
                return true;
             case SDLK_F9:
             case SDLK_F9 + ss_Alt:
                // ship and planet comments are handled through the script interface
                tryEditMarkerComment(loc);
                return true;
             case 'm' + ss_Ctrl:
                createCannedMarker(event.key.arg, loc);
                return true;
             case 'm':
                if (event.key.arg != 0) {
                    createCannedMarker(event.key.arg, loc);
                } else if (int mf = global_minefield_type.findNearest(loc.getLocation())) {
                    global_minefield_type.setCurrent(mf);
                    doMinefieldScreen();
                } else {
                    WMessageChartMode::showMessage(*getParent(), _("No Minefields"));
                }
                return true;
             case 'u':
                if (int ufo = global_ufo_type.findNearest(loc.getLocation())) {
                    global_ufo_type.setCurrent(ufo);
                    doUfoScreen();
                } else {
                    WMessageChartMode::showMessage(*getParent(), _("No Ufos"));
                }
                return true;
             case 'i':
                if (int ion = global_ion_type.findNearest(loc.getLocation())) {
                    global_ion_type.setCurrent(ion);
                    doIonStormScreen();
                } else {
                    WMessageChartMode::showMessage(*getParent(), _("No Ion Storms"));
                }
                return true;
             case 'c':
                doColorChange(loc);
                return true;
             case 'v':
                doMoveMarker(loc, *getParent());
                return true;
             case 'p':
                doDraw(loc, *getParent());
                return true;
             case 'd':
                doDistance(loc, *getParent());
                return true;
             case 'r':
                doMarkRange(loc, *getParent());
                return true;
             case 'g':
                doGoTo(loc);
                return true;
             case SDLK_DELETE:
                doDelete(loc, *getParent());
                return true;
             case ss_Alt + 'l':
                editLabelExpressions();
                loc.markChanged();     // Force redraw
                return true;
             case ss_Alt + 'o':
                doChartConfig();
                loc.markChanged();     // Force redraw
                return true;
             case 'x':
                if (chart_conf.getMode() == GChartConfiguration::Circular) {
                    GPoint pt = chart_conf.getCanonicalLocation(loc.getLocation());
                    GPoint pt2;
                    if (pt != loc.getLocation()) {
                        loc.setLocation(pt);
                    } else if (chart_conf.getPointAlias(pt, pt2, 1, true)) {
                        loc.setLocation(pt2);
                    }
                }
                return true;
             default:
                if ((event.key.code & ss_Alt) != 0) {
                    const GChartViewport& vp = loc.getViewport();
                    if (getChartOpts(false, vp.mult, vp.divi).toggleOptionKey(event.key.code & ~(ss_Alt | ss_Ctrl))) {
                        loc.markChanged();
                        return true;
                    }
                }
                break;
            }
        }
    }
    return false;
}
