/**
  *  \file client/lru.cc
  *  \brief Least-Recently-Used (and Predefined) Lists
  *
  *  This module provides data structures to manage LRU lists and
  *  lists of predefined items. The same data structure is used for
  *  both, allowing easier combination of both in one widget.
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/lru.h"

static const uint32_t LRU_LIMIT = 5;

/** Find starting point for comparison. If \c s starts with a
    bracketed item, returns the position past that. */
static string_t::size_type
findStart(const string_t& s)
{
    string_t::size_type n;
    if (s.size() > 0 && s[0] == '[' && (n = s.find(']')) != string_t::npos) {
        do {
            ++n;
        } while (n < s.size() && s[n] == ' ');
        return n;
    } else {
        return 0;
    }
}

/******************************** TLRUList *******************************/

/** Create empty list. */
LRUList::LRUList()
{ }

/** Destructor. */
LRUList::~LRUList()
{ }

/** Find value. Honors the bracket convention.
    \param value Value to search
    \return Index [0,size()) if found, size() if not found */
uint32_t
LRUList::findValue(const string_t& value) const
{
    uint32_t i = 0;
    while (i < items.size() && !isSameValue(value, items[i]->getValue())) {
        ++i;
    }
    return i;
}

/** Append new item.
    \param item Newly-allocated item. Will be owned by the list. */
void
LRUList::pushNew(LRUListItem* item)
{
    items.push_back(item);
}

/** Add new item, with LRU-style limiting. The item will be added at
    the front. Duplicates will be removed (this will make it appear to
    move the item from its original position to the front, but in case
    the items differ in their bracket, the new bracket will be used). 
    The list will be limited to LRU_LIMIT items. */
void
LRUList::pushNewLimit(LRUListItem* item)
{
    // add an item and move it to the front
    items.push_back(item);
    moveToFront(items.size()-1);

    // remove duplicates
    uint32_t in = 1, out = 1;
    while (in < items.size() && out < LRU_LIMIT) {
        if (isSameValue(items[in]->getName(), item->getName())) {
            // skip
            ++in;
        } else {
            // keep
            if (in != out) {
                items.swap_elems(in, out);
            }
            ++in, ++out;
        }
    }

    // drop unused items
    while (items.size() > out) {
        items.pop_back();
    }
}

/** Move item to front, given an index.
    \param index [0,size()) Item to move to front. */
void
LRUList::moveToFront(uint32_t index)
{
    if (index < items.size()) {
        while (index > 0) {
            items.swap_elems(index, index-1);
            --index;
        }
    }
}

/** Clear this list. */
void
LRUList::clear()
{
    items.clear();
}

/** Compare two values, honoring bracket convention.
    \param a,b Items to compare
    \return true iff both are considered equivalent */
bool
LRUList::isSameValue(const string_t& a, const string_t& b)
{
    string_t::size_type as = findStart(a);
    string_t::size_type bs = findStart(b);
    string_t::size_type an = a.size() - as;
    string_t::size_type bn = b.size() - bs;

    return an == bn && a.compare(as, an, b, bs, bn) == 0;
}
