/**
  *  \file flak/flak-db.cc
  *  \brief FLAK VCR Database Implementation
  *
  *  This implements a GVcrDatabase descendant for FLAK. Because GVcrDatabase
  *  is a PCC2 client core thing, and (indirectly) refers other PCC2
  *  infrastructure, this module should not be linked into programs using
  *  the PDK (and the PDK mini-platform).
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include "flak/flak-db.h"
#include "cpluslib/misc.h"
#include "game/config.h"
#include "io/io.h"
#include "io/streams.h"
#include "util/except.h"
#include "util/misc.h"
#include "util/nls.h"
#include "util/progress.h"
#include "util/recode.h"

static void
copyBefore(GVcrObject& obj, const FlakShip& src)
{
    obj.setMass(src.data.mass);
    obj.setShield(src.data.shield_init);
    obj.setDamage(src.data.damage_init);
    obj.setCrew(src.data.crew_init);
    obj.setId(src.data.id);
    obj.setOwner(src.data.player);
    // obj.setRace - not required
    obj.setPicture(1);
    obj.setHull(src.data.hull);
    obj.setBeamType(src.data.beam_type);
    obj.setBeamCount(src.data.beam_count);
    obj.setTorpType(src.data.torp_type);
    obj.setTorpCount(src.data.torp_count);
    obj.setTorpLauncherCount(src.data.torp_lcount);
    obj.setBayCount(src.data.bay_count);
    obj.setFighterCount(src.data.fighter_count);
    obj.setExperienceLevel(src.data.level);
    obj.setIsPlanet((src.data.flags & flak_IsPlanet) != 0);
    obj.setName(convertUtf8ToGame(getBasicString(src.data.name)));
}

static void
copyAfter(GVcrObject& obj, const FlakShip& src)
{
    // Start by copying before, then patch in result
    copyBefore(obj, src);
    obj.setShield(int(src.status.shield + 0.5));
    obj.setDamage(int(src.status.damage + 0.5));
    obj.setCrew(int(src.status.crew + 0.5));
    obj.setTorpCount(src.status.torp_count);
    obj.setFighterCount(src.status.fighter_count);
}

/****************************** FlakVcrEntry *****************************/

/** Constructor. Initialize this VCR entry from data from file.
    \param data [in] Battle data, starting with length word. */
FlakVcrEntry::FlakVcrEntry(string_t name, const Buffer<char>& data)
    : before(),
      after(),
      have_after(false)
{
    battle.loadFromStream(name, data, header);
    initBefore();
}

FlakVcrEntry::FlakVcrEntry(const FlakBattle& the_battle)
    : battle(the_battle),
      before(),
      after(),
      have_after(false)
{
    zeroFill(header);
    initBefore();
}

FlakVcrEntry::~FlakVcrEntry()
{ }

int
FlakVcrEntry::getNumObjects() const
{
    return battle.getNumShips();
}

GVcrObject&
FlakVcrEntry::getObject(int slot, bool after)
{
    if (after) {
        if (!have_after) {
            ProgressMonitor pm;
            prepareResult(NeedCompleteResult, pm);
        }
        return *this->after[slot];
    } else {
        return *this->before[slot];
    }
}

int
FlakVcrEntry::getOutcome(int slot)
{
    return battle.getShipByNumber(slot).data.ending_status;
}

void
FlakVcrEntry::prepareResult(int resultLevel, ProgressMonitor& pm)
{
    if ((resultLevel & ~NeedQuickOutcome) != 0 && !have_after) {
        // Play the fight
        if (battle.getNumFleets() != 0) {
            FlakNullVisualizer nv(battle);
            battle.init();
            int tick = 0;
            while (!battle.playOneTick()) {
                if (++tick > 100) {
                    pm.tick();
                    tick = 0;
                }
            }
        }

        // Build the result
        for (int i = 0, n = battle.getNumShips(); i < n; ++i) {
            after.push_back(new GVcrObject());
            copyAfter(*after.back(), battle.getShipByNumber(i));
        }
        have_after = true;
    }
}

GVcrEntry::Playability
FlakVcrEntry::getPlayability()
{
    return IsPlayable;
}

string_t
FlakVcrEntry::getAlgorithmName() const
{
    return _("FLAK");
}

bool
FlakVcrEntry::isESBActive()
{
    return config.AllowEngineShieldBonus();
}

void
FlakVcrEntry::initBefore()
{
    for (int i = 0, n = battle.getNumShips(); i < n; ++i) {
        before.push_back(new GVcrObject());
        copyBefore(*before.back(), battle.getShipByNumber(i));
    }
}

/**************************** FlakVcrDatabase ****************************/

/** Constructor. Create empty FLAK VCR Database. */
GFlakVcrDatabase::GFlakVcrDatabase()
{ }

/** Constructor. Create database by reading from file. */
GFlakVcrDatabase::GFlakVcrDatabase(Stream& s)
{
    loadFromFile(s);
}

/** Destructor. */
GFlakVcrDatabase::~GFlakVcrDatabase()
{ }

/** Read one battle from given stream.
    \return true if successful, false on error. */
void
GFlakVcrDatabase::readOneBattle(Stream& s)
{
    Buffer<char> data;
    data.ensureSize(4);
    s.readT(data.ptr(), 4);
    uint32 size = getUint32(data.ptr());
    data.ensureSize(size);
    s.readT(data.ptr()+4, size-4);
    battles.push_back(new FlakVcrEntry(s.getName(), data));
}

/** Load battles from a file. Battles will be appended to this
    database. */
void
GFlakVcrDatabase::loadFromFile(Stream& s)
{
    getStructureT(s, last_header);
    if (std::memcmp(last_header.magic, FLAK_MAGIC, sizeof(FLAK_MAGIC)) != 0)
        throw FileFormatException(s, _("File is missing required signature"));
    if (last_header.filefmt_version != 0)
        throw FileFormatException(s, _("Unsupported file format version"));

    for (int i = 0; i < last_header.num_battles; ++i) {
        readOneBattle(s);
    }
}

void
GFlakVcrDatabase::addBattle(const FlakBattle& b)
{
    battles.push_back(new FlakVcrEntry(b));
}

GTimestamp
GFlakVcrDatabase::getTimestamp()
{
    return GTimestamp(last_header.timestamp);
}

int
GFlakVcrDatabase::getNumBattles() const
{
    return battles.size();
}

FlakVcrEntry&
GFlakVcrDatabase::getBattle(int nr)
{
    return *battles[nr];
}
