/**
  *  \file game/mine.h
  *  \brief Minefields
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_MINE_H
#define PCC_V2_GAME_MINE_H

#include <map>
#include "game/obj.h"
#include "game/objtypes.h"
#include "cpluslib/types.h"
#include "cpluslib/iterator.h"

class GUniverse;
class GMinefieldList;

//! A Minefield.
class GMinefield : public GCircularObject {
    // friend class GMinefieldList;

    /* Data of last scan */
    int   id;
    int   x, y;
    int   owner;
    bool  web;
    int32 units;
    int   turn;

    int   last_action;

    /* Previous data */
    int32 units_prev;
    int   turn_prev;

    /* Cached current values */
    int   radius_now;
    int32 units_now;
 public:
    GMinefield(int id, int ac = 0);

    void internalCheck(const GUniverse& univ);

    void setData(int x, int y, int own, int32 units, int turn);
    void setData(int x, int y, int own, int32 units, int turn, bool web);

    /* GMapObject interface */
    string_t getName() const;
    string_t getFullName() const;
    int      getId() const;
    int      getOwner() const;
    bool     isExistant() const;
    GPoint   getPos() const;
    int      getRadius() const;
    int32_t  getRadiusSquared() const;

    /* Minefield Interface */
    int32 getUnitsAfterDecay(int32 orig_units) const;
    static int getRadiusFromUnits(int32 units);

    double getPassRate(double distance, bool cloaked, int player) const;

    int getAction() const
        { return last_action; }
    void setAction(int ac)
        { last_action = ac; }
    int32 getUnits() const
        { return units_now; }
    bool isWeb() const
        { return web; }
    int getTurnLastSeen() const
        { return turn; }
    int32 getUnitsLastSeen() const
        { return units; }
    int32 getUnitsForLaying() const;

    /** Actions that lead to Minefield Update. These must keep their
        relative and absolute values, because these equal the `ScanType'
        of a UTILx.DAT entry, plus one. */
    enum {
        mfa_None,               ///< No action, i.e. minefield not scanned this turn
        mfa_Laid,               ///< Minefield laid this turn.
        mfa_Swept,              ///< Minefield swept this turn.
        mfa_Scanned             ///< Minefield scanned this turn.
    };
};

class GMinefieldContainer {
    typedef std::map<int,GMinefield*> minemap_t;
    minemap_t mines;
 public:
    typedef member_iterator<minemap_t::iterator,
                            GMinefield*,
                            &minemap_t::value_type::second> ptr_iterator;
    typedef member_iterator<minemap_t::const_iterator,
                            GMinefield*const,
                            &minemap_t::value_type::second> const_ptr_iterator;
    typedef dereferencing_iterator<ptr_iterator> iterator;
    typedef dereferencing_iterator<const_ptr_iterator> const_iterator;

    GMinefieldContainer();
    ~GMinefieldContainer();

    iterator begin();
    iterator end();
    const_iterator begin() const;
    const_iterator end() const;

    iterator erase(iterator i);
    void     erase(int i);

    void clear();

    GMinefield& getCreate(int id, int cause);
    GMinefield* get(int id) const;

    Signal1<int> sig_erase;
};

class GMinefieldType : public GIndexedObjectType {
    SignalConnection conn_erase;
    void onErase(int id);

 public:
    GMinefieldType();

    void init();

    bool isExistant(int index);
    GMinefield& getObject(int index);
    GMinefield& getMinefield()
        { return getObject(getCurrent()); }
};



inline GMinefieldContainer::iterator
GMinefieldContainer::begin()
{
    return ptr_iterator(mines.begin());
}

inline GMinefieldContainer::iterator
GMinefieldContainer::end()
{
    return ptr_iterator(mines.end());
}

inline GMinefieldContainer::const_iterator
GMinefieldContainer::begin() const
{
    return const_ptr_iterator(mines.begin());
}

inline GMinefieldContainer::const_iterator
GMinefieldContainer::end() const
{
    return const_ptr_iterator(mines.end());
}

#endif
