/**
  *  \file int/if/ionif.cc
  *  \brief Interpreter Game Interface: Ion Storms
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/if/ionif.h"
#include "cpluslib/format.h"
#include "cpluslib/misc.h"
#include "game/global.h"
#include "game/ion.h"
#include "game/tables.h"
#include "int/compound.h"
#include "int/if/ifutil.h"
#include "int/interror.h"
#include "int/value.h"
#include "int/args.h"
#include "int/contextio.h"
#include "int/if/objif.h"
#include "int/propenum.h"

/** Implementation of "Storm(nn)".
    \param args function parameters
    \return value representing "Storm(nn)" */
IntValue*
IFIonGet(IntArgBlock& args)
{
    int32_t iid;
    args.checkArgumentCount(1);
    if (!checkIntArg(iid, args.getNext(), 1, NUM_STORMS))
        return 0;

    if (!haveDisplayedTurn())
        return 0;

    return new IntIonContext(iid);
}

/** Implementation of "ForEach Storm".
    \return Context representing first ion storm */
IntContext*
IFIonMake()
{
    if (!haveDisplayedTurn())
        return 0;

    int iid = global_ion_type.getNextNoWrap(0, false);
    if (iid != 0)
        return new IntIonContext(iid);
    else
        return 0;
}

/***************************** IntIonContext *****************************/

namespace {
    enum { IonProperty, IonMethod };
}

static const TIntMapping ion_mapping[] = {
    { "CLASS",     iipClass,       IonProperty, thInt },
    { "HEADING",   iipHeadingName, IonProperty, thString },
    { "HEADING$",  iipHeadingInt,  IonProperty, thInt },
    { "ID",        iipId,          IonProperty, thInt },
    { "LOC.X",     iipLocX,        IonProperty, thInt },
    { "LOC.Y",     iipLocY,        IonProperty, thInt },
    { "MARK",      0,              IonMethod,   thProcedure },
    { "MARKED",    iipMarked,      IonProperty, thBool },
    { "NAME",      iipName,        IonProperty, thString },
    { "RADIUS",    iipRadius,      IonProperty, thInt },
    { "SPEED",     iipSpeedName,   IonProperty, thString },
    { "SPEED$",    iipSpeedInt,    IonProperty, thInt },
    { "STATUS",    iipStatusName,  IonProperty, thString },
    { "STATUS$",   iipStatusFlag,  IonProperty, thBool },
    { "UNMARK",    1,              IonMethod,   thProcedure },
    { "VOLTAGE",   iipVoltage,     IonProperty, thInt },
};

static const IntObjectProcedureValue::call_t ion_methods[] = {
    IFObjMark,                  // 0
    IFObjUnmark,                // 1
};


/** Constructor.
    \param iid Ion Storm Id */
IntIonContext::IntIonContext(int iid)
    : iid(iid)
{ }

/** Destructor. */
IntIonContext::~IntIonContext()
{ }

int32_t
IntIonContext::lookup(const string_t& name)
{
    return lookupName(name, ion_mapping, countof(ion_mapping));
}

void
IntIonContext::set(int32_t index, IntValue* value)
{
    switch (ion_mapping[index].domain) {
     case IonProperty:
        setIonProperty(global_ion_type.getObject(iid), IntIonProperty(ion_mapping[index].index), value);
        break;
     default:
        throw IntError::notAssignable();
    }
}

IntValue*
IntIonContext::get(int32_t index)
{
    switch (ion_mapping[index].domain) {
     case IonProperty:
        return getIonProperty(global_ion_type.getObject(iid), IntIonProperty(ion_mapping[index].index));
     case IonMethod:
        return new IntObjectProcedureValue(global_ion_type.getObject(iid), ion_methods[ion_mapping[index].index]);
     default:
        return 0;
    }
}

bool
IntIonContext::next()
{
    int n = global_ion_type.getNextNoWrap(iid, false);
    if (n != 0) {
        iid = n;
        return true;
    } else {
        return false;
    }
}

IntContext*
IntIonContext::clone()
{
    return new IntIonContext(iid);
}

GObject*
IntIonContext::getObject()
{
    return &global_ion_type.getObject(iid);
}

string_t
IntIonContext::toString(bool /*readable*/)
{
    return format("Storm(%d)", iid);
}

void
IntIonContext::store(IntTagNode& sv, Stream& /*aux*/)
{
    sv.tag   = Tag_Ion;
    sv.value = iid;
}

void
IntIonContext::enumProperties(IntPropertyAcceptor& acceptor)
{
    acceptor.enumMapping(ion_mapping, countof(ion_mapping));
}


/** Get Ion Storm property.
    \param ion Storm to query
    \param iip Property to query
    \return property value */
IntValue*
getIonProperty(const GIonStorm& ion, IntIonProperty iip)
{
    if (!ion.isExistant())
        return 0;
    switch (iip) {
     case iipClass:
        return makeIntValue(ion.getClass());
     case iipHeadingInt:
        return makeIntValue(ion.getHeading());
     case iipHeadingName:
        return makeStringValue(getHeadingName(ion.getHeading()));
     case iipId:
        return makeIntValue(ion.getId());
     case iipLocX:
        return makeIntValue(ion.getPos().x);
     case iipLocY:
        return makeIntValue(ion.getPos().y);
     case iipMarked:
        return makeBoolValue(ion.isMarked());
     case iipName:
        return makeStringValue(ion.getName());
     case iipRadius:
        return makeIntValue(ion.getRadius());
     case iipSpeedInt:
        return makeIntValue(ion.getSpeed());
     case iipSpeedName:
        return makeStringValue(format("Warp %d", ion.getSpeed()));
     case iipStatusFlag:
        return makeBoolValue(ion.isGrowing());
     case iipStatusName:
        return makeStringValue(ion.isGrowing() ? "Growing" : "Weakening");
     case iipVoltage:
        return makeIntValue(ion.getVoltage());
    }
    return 0;
}

/** Set Ion Storm property. */
void
setIonProperty(GIonStorm& /*ion*/, IntIonProperty /*iip*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}
