/**
  *  \file int/statement.h
  *  \brief Interpreter: Compiling statements
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_INT_STATEMENT_H
#define PCC_V2_INT_STATEMENT_H

#include "int/expr.h"
#include "int/csource.h"
#include "cpluslib/ptrvector.h"

class IntExecutionContext;

/** Statement compilation context. In addition to some flags, this also manages
    an execution context and code generation for Break/Continue/Return.

    Despite having virtual functions, this class has no virtual destructor.
    Objects of its type are not allocated dynamically. */
class IntStatementCompilationContext : public IntCompilationContext {
 public:
    IntStatementCompilationContext();
    IntStatementCompilationContext(const IntStatementCompilationContext* parent);

    virtual void compileBreak(IntBytecodeObject& bco) const;
    virtual void compileContinue(IntBytecodeObject& bco) const;
    virtual void compileCleanup(IntBytecodeObject& bco) const;

    IntStatementCompilationContext& withFlags(uint32_t flag);
    IntStatementCompilationContext& withoutFlags(uint32_t flag);
    IntStatementCompilationContext& withExecutionContext(IntExecutionContext* exec_context);
    IntStatementCompilationContext& setOneLineSyntax();
    IntStatementCompilationContext& setBlockSyntax();

    IntExecutionContext* getExecutionContext() const;

 private:
    /** Parent command compilation context. Command compilation contexts are chained
        when nested blocks are used. */
    const IntStatementCompilationContext* parent;

    /** Execution context. When this is set, keywords are resolved using its context.
        When this is not set, we generate generic code, which has 'Eval' statements
        at ambiguous places. */
    IntExecutionContext* exec_context;
};

/** Statement compiler. Takes an IntCommandSource, and compiles it into an
    IntBytecodeObject using a IntStatementCompilationContext.

    Main entry points are compile() for a single statement, and compileList() for
    a statement sequence. Those call each other recursively for nested statements.

    Acceptance of one-line and block commands is configured using the
    IntStatementCompilationContext.

    Errors are reported by throwing IntError. */
class IntStatementCompiler {
 public:
    enum StatementResult {
        EndOfInput,              ///< End of input reached. Exit. Only if not WantTerminators.
        Terminator,              ///< Terminator statement left in current token. Only if WantTerminators.
        CompiledStatement,       ///< Successfully compiled a single-line statement.
        CompiledBlock,           ///< Successfully compiled a multi-line statement. Only if not RefuseBlocks.
        CompiledExpression       ///< Successfully compiled an expression. Its result remains on stack. Only if not ExpressionsAreStatements.
    };

    IntStatementCompiler(IntCommandSource& cs);

    StatementResult compile(IntBytecodeObject& bco, const IntStatementCompilationContext& scc);
    StatementResult compileList(IntBytecodeObject& bco, const IntStatementCompilationContext& scc);

 private:
    StatementResult compileAmbiguousStatement(IntBytecodeObject& bco, const IntStatementCompilationContext& scc);
    StatementResult compileAmbiguousSingleWord(const string_t& name, IntBytecodeObject& bco, const IntStatementCompilationContext& scc);
    StatementResult compileAmbiguousRuntimeSwitch(const string_t& name, IntBytecodeObject& bco, const IntStatementCompilationContext& scc);
    StatementResult compileAbort(IntBytecodeObject& bco, const IntStatementCompilationContext& scc);
    StatementResult compileBind(IntBytecodeObject& bco, const IntStatementCompilationContext& scc);
    StatementResult compileCreateKeymap(IntBytecodeObject& bco, const IntStatementCompilationContext& scc);
    StatementResult compileCreateProperty(IntBytecodeObject& bco, const IntStatementCompilationContext& scc, IntOpcode::Special minor, const char* prefix);
    StatementResult compileDim(IntBytecodeObject& bco, const IntStatementCompilationContext& scc, IntOpcode::Scope scope, int num_scopes);
    StatementResult compileDo(IntBytecodeObject& bco, const IntStatementCompilationContext& scc);
    StatementResult compileEval(IntBytecodeObject& bco, const IntStatementCompilationContext& scc);
    StatementResult compileFor(IntBytecodeObject& bco, const IntStatementCompilationContext& scc);
    StatementResult compileForEach(IntBytecodeObject& bco, const IntStatementCompilationContext& scc);
    StatementResult compileIf(IntBytecodeObject& bco, const IntStatementCompilationContext& scc);
    StatementResult compileLoad(IntBytecodeObject& bco, const IntStatementCompilationContext& scc, bool mustSucceed);
    StatementResult compileOn(IntBytecodeObject& bco, const IntStatementCompilationContext& scc);
    StatementResult compileOption(IntBytecodeObject& bco, const IntStatementCompilationContext& scc);
    StatementResult compilePrint(IntBytecodeObject& bco, const IntStatementCompilationContext& scc);
    StatementResult compileReturn(IntBytecodeObject& bco, const IntStatementCompilationContext& scc);
    StatementResult compileRunHook(IntBytecodeObject& bco, const IntStatementCompilationContext& scc);
    StatementResult compileSelect(IntBytecodeObject& bco, const IntStatementCompilationContext& scc);
    StatementResult compileSelectionExec(IntBytecodeObject& bco, const IntStatementCompilationContext& scc);
    StatementResult compileSub(IntBytecodeObject& bco, const IntStatementCompilationContext& scc, bool proc);
    StatementResult compileStruct(IntBytecodeObject& bco, const IntStatementCompilationContext& scc);
    StatementResult compileTry(IntBytecodeObject& bco, const IntStatementCompilationContext& scc);
    StatementResult compileWith(IntBytecodeObject& bco, const IntStatementCompilationContext& scc);
    StatementResult compileExpressionStatement(IntBytecodeObject& bco, const IntStatementCompilationContext& scc);
    StatementResult compileProcedureCall(IntBytecodeObject& bco, const IntStatementCompilationContext& scc);
    bool compileInitializer(IntBytecodeObject& bco, const IntStatementCompilationContext& scc);
    bool compileTypeInitializer(IntBytecodeObject& bco, const IntStatementCompilationContext& scc, const string_t& typeName);
    void compileArgumentExpression(IntBytecodeObject& bco, const IntStatementCompilationContext& scc);
    void compileArgumentCondition(IntBytecodeObject& bco, const IntStatementCompilationContext& scc, IntBytecodeObject::label_t ift, IntBytecodeObject::label_t iff);
    void compileSelectCondition(IntBytecodeObject& bco, const IntStatementCompilationContext& scc, IntBytecodeObject::label_t ldo);
    void parseArgumentList(ptr_vector<IntExprNode>& args);
    void parseEndOfLine();
    void validateName(const string_t& name);
    
    IntCommandSource& cs;
};

void parseCommandArgumentList(IntTokenizer& tok, ptr_vector<IntExprNode>& args);

/**************************** Inline Functions ***************************/

/** Add some flags.
    \return *this */
inline IntStatementCompilationContext&
IntStatementCompilationContext::withFlags(uint32_t flag)
{
    IntCompilationContext::withFlags(flag);
    return *this;
}

/** Remove some flags.
    \return *this */
inline IntStatementCompilationContext&
IntStatementCompilationContext::withoutFlags(uint32_t flag)
{
    IntCompilationContext::withoutFlags(flag);
    return *this;
}

/** Set execution context. The execution context is used to resolve ambiguous statements.
    It is set to a non-null value if and only if the compiled statement is a one-line
    statement going to be executed in that context, where the context does not yet contain
    a frame for the BCO we're compiling into.

    If the execution context is not set, some statements (see compileAmbiguousStatement)
    cannot be compiled and are turned into "evals" instructions instead.

    Rationale: it must not be set if the script can change the context, which is the case
    when we have anything that can follow a "Sub" or "Dim". Hence, only one-liners which
    cannot have anything that follows. It must be set, however, when executing the one-liner
    that results fromt he "evals" instruction, so we can guarantee termination.

    \return *this */
inline IntStatementCompilationContext&
IntStatementCompilationContext::withExecutionContext(IntExecutionContext* exec_context)
{
    this->exec_context = exec_context;
    return *this;
}

/** Set flags for one-line statement syntax.
    - Add RefuseBlocks because we're a one-liner
    - Add ExpressionsAreStatements because we're a substatement
    - Remove WantTerminators to reject code like "If a then EndSub"
    \return *this*/
inline IntStatementCompilationContext&
IntStatementCompilationContext::setOneLineSyntax()
{
    return withFlags(RefuseBlocks | ExpressionsAreStatements).withoutFlags(WantTerminators);
}

/** Set flags for block statement syntax.
    - Remove RefuseBlocks because we're a block (should already be clear)
    - Add ExpressionsAreStatements because we're a substatement
    - Add WantTerminators to find end of block"
    \return *this */
inline IntStatementCompilationContext&
IntStatementCompilationContext::setBlockSyntax()
{
    return withoutFlags(RefuseBlocks).withFlags(WantTerminators | ExpressionsAreStatements);
}

/** Get execution context. */
inline IntExecutionContext*
IntStatementCompilationContext::getExecutionContext() const
{
    return exec_context;
}

#endif
