/**
  *  \file u/t_io_textfile.cc
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "io/textfile.h"

#include "u/t_io.h"
#include "io/memstream.h"

void
IoTextfileTestSuite::testGetChar()
{
    MemoryStream ms("123\012abc");
    TextFile tf(ms);

    TS_ASSERT_EQUALS(tf.peekByte(), '1');
    TS_ASSERT_EQUALS(tf.getByte(),  '1');

    TS_ASSERT_EQUALS(tf.peekByte(), '2');
    TS_ASSERT_EQUALS(tf.getByte(),  '2');

    TS_ASSERT_EQUALS(tf.peekByte(), '3');
    TS_ASSERT_EQUALS(tf.getByte(),  '3');

    TS_ASSERT_EQUALS(tf.peekByte(), 10);
    TS_ASSERT_EQUALS(tf.getByte(),  10);

    TS_ASSERT_EQUALS(tf.peekByte(), 'a');
    TS_ASSERT_EQUALS(tf.getByte(),  'a');

    TS_ASSERT_EQUALS(tf.peekByte(), 'b');
    TS_ASSERT_EQUALS(tf.getByte(),  'b');

    TS_ASSERT_EQUALS(tf.peekByte(), 'c');
    TS_ASSERT_EQUALS(tf.getByte(),  'c');

    TS_ASSERT_EQUALS(tf.peekByte(), -1);
    TS_ASSERT_EQUALS(tf.getByte(),  -1);

    TS_ASSERT_EQUALS(tf.peekByte(), -1);
    TS_ASSERT_EQUALS(tf.getByte(),  -1);

    TS_ASSERT_EQUALS(tf.peekByte(), -1);
    TS_ASSERT_EQUALS(tf.getByte(),  -1);
}

void
IoTextfileTestSuite::doGetLineTest(const char* data, const char* name)
{
    MemoryStream ms(data);
    TextFile tf(ms);
    string_t line;
    TSM_ASSERT(name, tf.getLine(line));
    TSM_ASSERT_EQUALS(name, line, "abc");

    TSM_ASSERT(name, tf.getLine(line));
    TSM_ASSERT_EQUALS(name, line, "def");

    TSM_ASSERT(name, !tf.getLine(line));
    TSM_ASSERT_EQUALS(name, line, "");
}

void
IoTextfileTestSuite::testGetLine()
{
    doGetLineTest("abc\012def\012",         "Unix");
    doGetLineTest("abc\015\012def\015\012", "DOS");
    doGetLineTest("abc\015def\015",         "Mac");
    doGetLineTest("abc\012def",             "Unix unterminated");
    doGetLineTest("abc\015\012def",         "DOS unterminated");
    doGetLineTest("abc\015def",             "Mac unterminated");

    doGetLineTest("abc\012def\012\032",         "Unix C-z");
    doGetLineTest("abc\015\012def\015\012\032", "DOS C-z");
    doGetLineTest("abc\015def\015\032",         "Mac C-z");
    doGetLineTest("abc\012def\032",             "Unix unterminated C-z");
    doGetLineTest("abc\015\012def\032",         "DOS unterminated C-z");
    doGetLineTest("abc\015def\032",             "Mac unterminated C-z");
}

void
IoTextfileTestSuite::testEncoding()
{
    {
        // Default encoding is latin1
        MemoryStream ms("\xC2\x80");
        TextFile tf(ms);
        string_t line;
        TS_ASSERT(tf.getLine(line));
        TS_ASSERT_EQUALS(line, "\xC3\x82\xC2\x80");
    }

    {
        // Explicitly-set latin1
        MemoryStream ms("\xC2\x80");
        TextFile tf(ms);
        string_t line;
        tf.setCharacterSet(CharacterSet::getCharsetByName("ISO-8859-1"));
        TS_ASSERT(tf.getLine(line));
        TS_ASSERT_EQUALS(line, "\xC3\x82\xC2\x80");
    }

    {
        // Explicitly-set KOI8-R
        MemoryStream ms("\xC2\x80");
        TextFile tf(ms);
        string_t line;
        tf.setCharacterSet(CharacterSet::getCharsetByName("koi8-r"));
        TS_ASSERT(tf.getLine(line));

        // 0xC2 -> U+0431,   0xD0 0xB1
        // 0x80 -> U+2500,   0xE2 0x94 0x80
        TS_ASSERT_EQUALS(line, "\xD0\xB1\xE2\x94\x80");
    }

    {
        // UTF-8
        MemoryStream ms("\xC2\x80");
        TextFile tf(ms);
        string_t line;
        tf.setUtf8();
        TS_ASSERT(tf.getLine(line));
        TS_ASSERT_EQUALS(line, "\xC2\x80");
    }

    {
        // implicit UTF-8
        MemoryStream ms("\xEF\xBB\xBF\xC2\x80");
        TextFile tf(ms);
        string_t line;
        TS_ASSERT(tf.getLine(line));
        TS_ASSERT_EQUALS(line, "\xC2\x80");
    }

    {
        // not UTF-8
        MemoryStream ms("a\xEF\xBB\xBF\xC2\x80");
        TextFile tf(ms);
        string_t line;
        TS_ASSERT(tf.getLine(line));

        // 0xEF = 0xC3 0xAF
        // 0xBB = 0xC2 0xBB
        // 0xBF = 0xC2 0xBF
        TS_ASSERT_EQUALS(line,
                         "a"
                         "\xC3\xAF\xC2\xBB\xC2\xBF"
                         "\xC3\x82\xC2\x80");
    }

    {
        // also not UTF-8
        MemoryStream ms("a\n\xEF\xBB\xBF\xC2\x80");
        TextFile tf(ms);
        string_t line;
        TS_ASSERT(tf.getLine(line));
        TS_ASSERT_EQUALS(line, "a");
        TS_ASSERT(tf.getLine(line));
        TS_ASSERT_EQUALS(line,
                         "\xC3\xAF\xC2\xBB\xC2\xBF"
                         "\xC3\x82\xC2\x80");
    }

    {
        // not UTF-8, too
        MemoryStream ms("\xEF\xBB\n\xBF\xC2\x80");
        TextFile tf(ms);
        string_t line;
        TS_ASSERT(tf.getLine(line));
        TS_ASSERT_EQUALS(line, "\xC3\xAF\xC2\xBB");
        TS_ASSERT(tf.getLine(line));
        TS_ASSERT_EQUALS(line,
                         "\xC2\xBF"
                         "\xC3\x82\xC2\x80");
    }

    {
        // implicit UTF-8 overrides explicitly-set character set
        MemoryStream ms("\xEF\xBB\xBF\xC2\x80");
        TextFile tf(ms);
        string_t line;
        tf.setCharacterSet(CharacterSet::getCharsetByName("koi8-r"));
        TS_ASSERT(tf.getLine(line));
        TS_ASSERT_EQUALS(line, "\xC2\x80");
    }
}

void
IoTextfileTestSuite::testWrite()
{
    {
        // Test writing with Latin-1 charset
        MemoryStream ms;
        TextFile tf(ms);
        tf.writeLine("hi");
        tf.writeLine("there\xC2\xA1");
        TS_ASSERT(tf.getSize() == 10 || tf.getSize() == 12);
        TS_ASSERT_EQUALS(tf.getPos(), tf.getSize());
        tf.sync();
        TS_ASSERT(tf.getSize() == 10 || tf.getSize() == 12);
        TS_ASSERT_EQUALS(tf.getPos(), tf.getSize());

        TS_ASSERT(   ms.getContentsAsString() == "hi\nthere\xA1\n"
                  || ms.getContentsAsString() == "hi\rthere\xA1\r"
                  || ms.getContentsAsString() == "hi\r\nthere\xA1\r\n");
    }

    {
        // Test writing with CP437 charset
        MemoryStream ms;
        TextFile tf(ms);
        tf.setCharacterSet(CharacterSet::getCharsetByName("cp437"));
        tf.writeLine("hi");
        tf.writeLine("there\xC2\xA1");
        TS_ASSERT(tf.getSize() == 10 || tf.getSize() == 12);
        TS_ASSERT_EQUALS(tf.getPos(), tf.getSize());
        tf.sync();
        TS_ASSERT(tf.getSize() == 10 || tf.getSize() == 12);
        TS_ASSERT_EQUALS(tf.getPos(), tf.getSize());

        TS_ASSERT(   ms.getContentsAsString() == "hi\nthere\xAD\n"
                  || ms.getContentsAsString() == "hi\rthere\xAD\r"
                  || ms.getContentsAsString() == "hi\r\nthere\xAD\r\n");
    }

    {
        // Test writing with UTF-8 charset
        MemoryStream ms;
        TextFile tf(ms);
        tf.setUtf8();
        tf.writeLine("hi");
        tf.writeLine("there\xC2\x80");
        TS_ASSERT(tf.getSize() == 11 || tf.getSize() == 13);
        TS_ASSERT_EQUALS(tf.getPos(), tf.getSize());
        tf.sync();
        TS_ASSERT(tf.getSize() == 11 || tf.getSize() == 13);
        TS_ASSERT_EQUALS(tf.getPos(), tf.getSize());

        TS_ASSERT(   ms.getContentsAsString() == "hi\nthere\xC2\x80\n"
                  || ms.getContentsAsString() == "hi\rthere\xC2\x80\r"
                  || ms.getContentsAsString() == "hi\r\nthere\xC2\x80\r\n");
    }

    {
        // Test writing and exceeding buffer size
        MemoryStream ms;
        TextFile tf(ms);
        for (int i = 0; i < 10000; ++i)
            tf.writeLine("hi");
        TS_ASSERT(tf.getPos() == 30000 || tf.getPos() == 40000);
        TS_ASSERT(tf.getPos() == tf.getSize());
        tf.sync();

        string_t result = ms.getContentsAsString();
        TS_ASSERT(result.length() == 30000 || result.length() == 40000);
        TS_ASSERT_EQUALS(result.substr(0, 2), "hi");
        TS_ASSERT(result[2] == '\r' || result[2] == '\n');
    }
}

void
IoTextfileTestSuite::testMixedIO()
{
    {
        MemoryStream ms("abcdefghijklmnopqrstuvwxyz");
        TextFile tf(ms);
        TS_ASSERT_EQUALS(tf.getPos(), 0);
        TS_ASSERT_EQUALS(tf.getSize(), 26);

        TS_ASSERT_EQUALS(tf.getByte(), 'a');
        TS_ASSERT_EQUALS(tf.getByte(), 'b');
        TS_ASSERT_EQUALS(tf.getPos(), 2);

        tf.writeBytes("CD", 2);
        TS_ASSERT_EQUALS(tf.getPos(), 4);

        TS_ASSERT_EQUALS(tf.getByte(), 'e');
        TS_ASSERT_EQUALS(tf.getByte(), 'f');

        tf.seek(3);
        TS_ASSERT_EQUALS(tf.getPos(), 3);
        TS_ASSERT_EQUALS(tf.getByte(), 'D');

        char tmp[5];
        TS_ASSERT_EQUALS(tf.readBytes(tmp, 5), 5);
        TS_ASSERT_SAME_DATA(tmp, "efghi", 5);

        tf.seek(24);
        TS_ASSERT_EQUALS(tf.readBytes(tmp, 5), 2);
        TS_ASSERT_SAME_DATA(tmp, "yzghi", 5);
    }
}
