/**
  *  \file ui/richhelp.cc
  *  \brief Rich Text Help Text Parser
  *
  *  This module contains a function loadHelpPage() and associated
  *  utilities to render (part of) an XML file into a RichDocument.
  *  This is used for displaying help files.
  *
  *  Note that we assume valid, well-formed XML.
  *
  *  \todo When given an input such as
  *     <b>foo</b> <tt>bar</tt>
  *  (with arbitrary space between the tags) it will not render a
  *  space between "foo" and "bar".
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "ui/richhelp.h"
#include "ui/richtext.h"
#include "util/xmlread.h"
#include "ui/richdocument.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "util/unicodechars.h"
#include "ui/uiglobal.h"
#include "util/misc.h"
#include "ui/skin.h"
#include "ui/richobj.h"
#include "resmgr/resmgr.h"
#include "io/memstream.h"

namespace {

    class RichParserBase {
     public:
        RichParserBase(XmlReader& rdr);

        XmlReader& rdr;
        XmlReader::Token current;

        void next();
        bool isOpen(const char* what);
        RichText parseText(bool keepFormat);
        RichText parseTextItem(bool keepFormat);
        void parseUnknown();
    };
    
    /** XML to RichDocument parser. This class contains status information
        and functions for parsing RichText into XML. In particular, it
        saves a token type to implement a 1-lookahead parser. */
    class RichHelpRenderer : public RichParserBase {
        RichDocument& doc;

        void flushLine(RichText& line);
        void parseBlock(int listLevel);
        void parseBulletList(int listLevel);
        void parseCountedList(int listLevel);
        void parseKeyList(int listLevel);
        void parseDefinitionList(int listLevel);
        void parseListItem(int listLevel);
        void parseImage();
        void parseTable();
        void parseTableLine(std::vector<int>& cellWidths, int align);
        void parseTableCell(std::vector<int>& cellWidths, ptr_vector<RichText>& cellText, std::vector<int>& cellAlign, int align);
        void parsePre();

     public:
        RichHelpRenderer(RichDocument& doc, XmlReader& rdr);

        void parseDocument();
    };

    /** XML to RichText parser. */
    class RichTextParser : public RichParserBase {
     public:
        RichTextParser(XmlReader& rdr);
        RichText parse();
    };
}

static const char*const bullets[] = {
    UTF_SQUARE_BULLET,
    UTF_BULLET,
    "+",
    "-",
    ".",
};

/** Append string to RichText. This collapses multiple spaces into
    one, and strips newlines.
    \param out [in/out] RichText to append to
    \param haveSpace [in/out] Space status tracking
    \param in [in] Text to append */
static void
appendText(RichText& out, bool& haveSpace, const string_t& in)
{
    static const char SPACE[] = " \r\n\t";
    string_t::size_type a = 0;
    while (a < in.size()) {
        if (haveSpace) {
            /* out ends with a space. Skip all space in input. */
            a = in.find_first_not_of(SPACE, a);
            haveSpace = false;
        } else {
            /* out does not end with a space, so look for spaces in /in/. */
            string_t::size_type e = in.find_first_of(SPACE, a);
            if (e == in.npos) {
                out.append(string_t(in, a));
            } else {
                out.append(string_t(in, a, e-a));
                out.append(" ");
                haveSpace = true;
            }
            a = e;
        }
    }
}

/** Parse a color name. */
static UIColor::Color
parseColorName(const string_t& name)
{
    static const struct {
        const char* name;
        UIColor::Color color;
    } colors[] = {
        { "static", UIColor::tc_Static },
        { "green",  UIColor::tc_Green },
        { "yellow", UIColor::tc_Yellow },
        { "red",    UIColor::tc_Red },
        { "white",  UIColor::tc_White },
        { "blue",   UIColor::tc_Blue },
        { "dim",    UIColor::tc_Faded },
    };
    for (std::size_t i = 0; i < countof(colors); ++i) {
        if (name == colors[i].name) {
            return colors[i].color;
        }
    }
    return UIColor::tc_Static;
}

/** Render a key specification. Documents contain \<kbd> tags for whole key strings, such
    as "Alt+A". This splits the key string into individual keys. */
static RichText
renderKeys(const string_t& name)
{
    RichText result;
    string_t::size_type p = 0;
    while (p < name.size()) {
        /* Delimiters:
           "-" and "+" for key combinations "Alt-K"
           "/" for alternatives "Up/Down"
           ",.;: " for punctuation "Up, Down" */
        string_t::size_type e = name.find_first_of("-+/,.;: ", p+1);
        if (e == string_t::npos) {
            /* String ends */
            result.append(RichText(string_t(name, p)).withAttribute(RichTextStyleAttribute::Key));
            break;
        }

        /* String does not end after key, but may have punctuation */
        result.append(RichText(string_t(name, p, e-p)).withAttribute(RichTextStyleAttribute::Key));
        p = e;
        if (name[p] == '.') {
            e = name.find_first_not_of(". ", p+1);
        } else {
            e = name.find_first_not_of(" ", p+1);
        }
        if (e == string_t::npos) {
            /* String ends with punctuation */
            result.append(string_t(name, p));
            break;
        }

        /* String does not end after punctuation. As a special case, if it's a
           normal dash, turn that into a (shorter) hyphen. */
        if (e == p+1 && name[p] == '-') {
            result.append(UTF_HYPHEN);
        } else {
            result.append(string_t(name, p, e-p));
        }
        p = e;
    }

    return result;
}

/** Render a script. This performs some simple syntax coloring on the script.
    \param in Input script
    \param out Result will be appended here */
static void
renderScript(const RichText& in, RichText& out)
{
    const string_t plain(in.getText());
    enum {
        Plain,
        DQuote,
        DQuoteBS,
        SQuote,
        QuoteEnd,
        Comment
    } state = Plain;
    string_t::size_type start = 0;
    for (string_t::size_type i = 0; i < plain.size(); ++i) {
        char c = plain[i];
        if (state == Plain) {
            if (c == '"') {
                state = DQuote;
            } else if (c == '\'') {
                state = SQuote;
            } else if (c == '%') {
                state = Comment;
            }
            if (state != Plain) {
                if (start != i) {
                    out.append(in.substr(start, i-start));
                }
                start = i;
            }
        } else if (state == DQuote) {
            if (c == '\\') {
                state = DQuoteBS;
            } else if (c == '"') {
                state = QuoteEnd;
            }
        } else if (state == DQuoteBS) {
            state = DQuote;
        } else if (state == SQuote) {
            if (c == '\'') {
                state = QuoteEnd;
            }
        } else if (state == QuoteEnd) {
            if (c == '"') {
                state = DQuote;
            } else if (c == '\'') {
                state = SQuote;
            } else if (c == '%') {
                state = Comment;
            } else {
                state = Plain;
            }
            if (state != DQuote && state != SQuote) {
                out.append(in.substr(start, i-start).withAttribute(new RichTextColorAttribute(UIColor::tc_Green)));
                start = i;
            }
        } else if (state == Comment) {
            if (c == '\n') {
                out.append(in.substr(start, i-start).withAttribute(new RichTextColorAttribute(UIColor::tc_Red)));
                start = i;
                state = Plain;
            }
        }
    }

    if (state == Comment) {
        out.append(in.substr(start).withAttribute(new RichTextColorAttribute(UIColor::tc_Red)));
    } else if (state == DQuote || state == DQuoteBS || state == QuoteEnd) {
        out.append(in.substr(start).withAttribute(new RichTextColorAttribute(UIColor::tc_Green)));
    } else {
        out.append(in.substr(start));
    }
}

/**************************** RichParserBase ***************************/

/** Constructor. */
inline
RichParserBase::RichParserBase(XmlReader& rdr)
    : rdr(rdr),
      current(rdr.Eof)
{ }

/** Advance to next tag. */
inline void
RichParserBase::next()
{
    current = rdr.read();
}

/** Check for opening tag. If that tag is found, skips it from the
    token stream.
    \param what Tag to check for (case sensitive!) */
bool
RichParserBase::isOpen(const char* what)
{
    if (current == rdr.TagStart && rdr.getTag() == what) {
        next();
        return true;
    } else {
        return false;
    }
}

/** Parse text element. */
RichText
RichParserBase::parseText(bool keepFormat)
{
    /* Text consists of a, b, em, tt, key, font, where a and font have parameters. */
    bool haveSpace = true;
    RichText result;
    while (1) {
        if (current == rdr.Eof || current == rdr.Error || current == rdr.TagEnd) {
            /* EOF or error stop parsing, or surrounding tag ends */
            break;
        } else if (current == rdr.TagAttribute) {
            /* Spurious attribute from surrounding tag */
            next();
        } else if (current == rdr.Text) {
            /* Text */
            if (keepFormat) {
                result.append(rdr.getValue());
            } else {
                appendText(result, haveSpace, rdr.getValue());
            }
            next();
        } else {
            result.append(parseTextItem(keepFormat));
            next();
            haveSpace = false;
        }
    }
    return result;
}

RichText
RichParserBase::parseTextItem(bool keepFormat)
{
    if (isOpen("a")) {
        string_t target;
        while (current == rdr.TagAttribute) {
            if (rdr.getName() == "href") {
                target = rdr.getValue();
            }
            next();
        }
        return parseText(keepFormat).withAttribute(new RichTextLinkAttribute(target));
    } else if (isOpen("b")) {
        return parseText(keepFormat).withAttribute(RichTextStyleAttribute::Bold);
    } else if (isOpen("em")) {
        // Should be italic, but we don't have an italic font yet.
        // return parseText(keepFormat).withAttribute(RichTextStyleAttribute::Italic);
        return parseText(keepFormat).withAttribute(RichTextStyleAttribute::Bold);
            ;
    } else if (isOpen("tt")) {
        return parseText(keepFormat).withAttribute(RichTextStyleAttribute::Fixed);
    } else if (isOpen("kbd") || isOpen("key")) {
        return renderKeys(parseText(keepFormat).getText());
    } else if (isOpen("font")) {
        UIColor::Color c = UIColor::tc_Static;
        while (current == rdr.TagAttribute) {
            if (rdr.getName() == "color") {
                c = parseColorName(rdr.getValue());
            }
            next();
        }
        return parseText(keepFormat).withAttribute(new RichTextColorAttribute(c));
    } else if (current == rdr.TagStart) {
        /* An opening tag we don't understand; parse its content */
        next();
        return parseText(keepFormat);
    } else {
        /* Completely unknown, skip it */
        next();
        return RichText();
    }
}

void
RichParserBase::parseUnknown()
{
    /* Parse some unknown tag. When we're here, 'current' is the opening tag. */
    int nesting = 0;
    do {
        if (current == rdr.TagStart) {
            ++nesting;
        }
        if (current == rdr.TagEnd) {
            --nesting;
        }
        next();
    } while (current != rdr.Eof && nesting != 0);
}

/**************************** RichHelpRenderer ***************************/

RichHelpRenderer::RichHelpRenderer(RichDocument& doc, XmlReader& rdr)
    : RichParserBase(rdr),
      doc(doc)
{ }

/** Output line to document if present. Clears it afterwards. */
void
RichHelpRenderer::flushLine(RichText& line)
{
    if (line.size()) {
        doc.add(line);
        doc.addNewline();
        line.clear();
    }
}

/** Parse a block-level element.
    - h1/h2/h3 headings
    - p paragraphs
    - ul/ol/kl/dl lists

    Upon call, we're looking at the opening tag.
    Upon exit, we're looking at the next element after the block element.

    \param listLevel list nesting level */
void
RichHelpRenderer::parseBlock(int listLevel)
{
    if (isOpen("h1")) {
        doc.add(parseText(false).withAttribute(RichTextStyleAttribute::Big).withAttribute(RichTextStyleAttribute::Bold));
        doc.addParagraph();
        next();
    } else if (isOpen("h2")) {
        doc.add(parseText(false).withAttribute(RichTextStyleAttribute::Bold));
        doc.addParagraph();
        next();
    } else if (isOpen("h3")) {
        doc.add(parseText(false).withAttribute(RichTextStyleAttribute::Underline));
        doc.addParagraph();
        next();
    } else if (isOpen("pre")) {
        parsePre();
        next();
    } else if (isOpen("p")) {
        /* This is almost a copy of parseText() */
        RichText line;
        bool haveSpace = true;
        while (1) {
            if (current == rdr.Eof || current == rdr.Error || current == rdr.TagEnd) {
                /* EOF or error stop parsing, or surrounding tag ends */
                break;
            } else if (current == rdr.TagAttribute) {
                /* Spurious attribute from surrounding tag */
                next();
            } else if (current == rdr.Text) {
                /* Text */
                appendText(line, haveSpace, rdr.getValue());
                next();
            } else if (isOpen("img")) {
                /* Image */
                doc.add(line);
                line.clear();
                parseImage();
                next();
            } else if (isOpen("br")) {
                /* Newline */
                flushLine(line);
                while (current != rdr.Eof && current != rdr.Error && current != rdr.TagEnd) {
                    next();
                }
                next();
                haveSpace = true;
            } else {
                line.append(parseTextItem(false));
                next();
                haveSpace = false;
            }
        }
        doc.add(line);
        doc.addParagraph();
        next();
    } else if (isOpen("ul")) {
        parseBulletList(listLevel);
        next();
    } else if (isOpen("ol")) {
        parseCountedList(listLevel);
        next();
    } else if (isOpen("dl")) {
        parseDefinitionList(listLevel);
        next();
    } else if (isOpen("table")) {
        parseTable();
        next();
    } else if (isOpen("kl")) {
        parseKeyList(listLevel);
        next();
    } else if (isOpen("img")) {
        parseImage();
        next();
    } else {
        parseUnknown();
    }
}

/** Parse bullet list.
    Upon call, we're looking at the first token of the content.
    Upon exit, we're looking at the closing tag. */
void
RichHelpRenderer::parseBulletList(int listLevel)
{
    while (1) {
        if (current == rdr.Eof || current == rdr.Error || current == rdr.TagEnd) {
            /* EOF or error stop parsing, or surrounding tag ends */
            break;
        } else if (isOpen("li")) {
            /* List item */
            const char* bullet = bullets[listLevel % countof(bullets)];
            int margin = doc.getLeftMargin();
            doc.addAt(margin, bullet);
            doc.add(" "); /* needed to make sure setLeftMargin does not indent the bullet! */
            doc.setLeftMargin(margin + fonts[FONT_NORMAL]->getTextWidth(string_t(bullet) + " "));
            parseListItem(listLevel+1);
            doc.setLeftMargin(margin);
            if (listLevel == 0) {
                doc.addNewline();
            }
            next();
        } else {
            /* Unknown, skip it */
            next();
        }
    }
}

/** Parse counted list. */
void
RichHelpRenderer::parseCountedList(int listLevel)
{
    const int em = font_ems[FONT_NORMAL] * 3/2;
    int counter = 0;
    while (1) {
        if (current == rdr.Eof || current == rdr.Error || current == rdr.TagEnd) {
            /* EOF or error stop parsing, or surrounding tag ends */
            break;
        } else if (isOpen("li")) {
            /* List item */
            ++counter;
            int margin = doc.getLeftMargin();
            doc.addAt(margin, numToString(counter));
            doc.add(". ");
            doc.setLeftMargin(margin + em);
            parseListItem(listLevel+1);
            doc.setLeftMargin(margin);
            if (listLevel == 0) {
                doc.addNewline();
            }
            next();
        } else {
            /* Unknown, skip it */
            next();
        }
    }
}

/** Parse definition list. */
void
RichHelpRenderer::parseDefinitionList(int listLevel)
{
    const int em = font_ems[FONT_NORMAL] * 3/2;
    while (1) {
        if (current == rdr.Eof || current == rdr.Error || current == rdr.TagEnd) {
            /* EOF or error stop parsing, or surrounding tag ends */
            break;
        } else if (isOpen("di")) {
            /* List item */
            string_t name;
            while (current == rdr.TagAttribute) {
                if (rdr.getName() == "term") {
                    name = rdr.getValue();
                }
                next();
            }
            int margin = doc.getLeftMargin();
            doc.addAt(margin, RichText(name).withAttribute(RichTextStyleAttribute::Bold));
            doc.add(" ");
            doc.setLeftMargin(margin + em);
            parseListItem(listLevel+1);
            doc.setLeftMargin(margin);
            if (listLevel == 0) {
                doc.addNewline();
            }
            next();
        } else {
            /* Unknown, skip it */
            next();
        }
    }
}

/** Parse key list. */
void
RichHelpRenderer::parseKeyList(int listLevel)
{
    const int em = font_ems[FONT_NORMAL] * 5;
    while (1) {
        if (current == rdr.Eof || current == rdr.Error || current == rdr.TagEnd) {
            /* EOF or error stop parsing, or surrounding tag ends */
            break;
        } else if (isOpen("ki")) {
            /* List item */
            string_t name;
            while (current == rdr.TagAttribute) {
                if (rdr.getName() == "key") {
                    name = rdr.getValue();
                }
                next();
            }
            int margin = doc.getLeftMargin();
            doc.addAt(margin, bullets[listLevel % countof(bullets)]);
            doc.add(" ");
            doc.add(renderKeys(name));
            doc.tabTo(margin + em);
            doc.setLeftMargin(margin + em);
            parseListItem(listLevel+1);
            doc.setLeftMargin(margin);
            next();
        } else {
            /* Unknown, skip it */
            next();
        }
    }
    if (listLevel == 0) {
        doc.addNewline();
    }
}

/** Parse list item.
    Upon call, we're looking at the first token of the content.
    Upon exit, we're looking at the closing tag of the list item. */
void
RichHelpRenderer::parseListItem(int listLevel)
{
    RichText line;
    bool haveSpace = true;
    bool hadImage = false;
    while (1) {
        if (current == rdr.Eof || current == rdr.Error || current == rdr.TagEnd) {
            /* EOF or error stop parsing, or surrounding tag ends */
            break;
        } else if (current == rdr.Text) {
            /* Text */
            appendText(line, haveSpace, rdr.getValue());
            next();
        } else if (isOpen("ul")) {
            /* Sublist */
            flushLine(line);
            parseBulletList(listLevel);
            next();
            haveSpace = true;
        } else if (isOpen("ol")) {
            /* Sublist */
            flushLine(line);
            parseCountedList(listLevel);
            next();
            haveSpace = true;
        } else if (isOpen("dl")) {
            /* Sublist */
            flushLine(line);
            parseDefinitionList(listLevel);
            next();
            haveSpace = true;
        } else if (isOpen("kl")) {
            /* Sublist */
            flushLine(line);
            parseKeyList(listLevel);
            next();
            haveSpace = true;
        } else if (isOpen("table")) {
            /* Table */
            flushLine(line);
            parseTable();
            next();
            haveSpace = true;
        } else if (isOpen("pre")) {
            /* Preformatted */
            parsePre();
            next();
            haveSpace = true;
        } else if (isOpen("img")) {
            /* Image */
            doc.add(line);
            line.clear();
            parseImage();
            next();
            hadImage = true;
        } else if (isOpen("br")) {
            /* Newline */
            flushLine(line);
            while (current != rdr.Eof && current != rdr.Error && current != rdr.TagEnd) {
                next();
            }
            next();
            haveSpace = true;
        } else {
            /* Formatted text */
            line.append(parseTextItem(false));
            next();
            haveSpace = false;
        }
    }

    if (line.size()) {
        doc.add(line);
        doc.addNewline();
    } else if (hadImage) {
        doc.addNewline();
    }
}

/** Parse image. Upon entry, the "img" tag has been read. Upon exit,
    we're looking at the closing tag. */
void
RichHelpRenderer::parseImage()
{
    /* Parse the tag */
    string_t img;
    string_t align;
    while (current == rdr.TagAttribute) {
        if (rdr.getName() == "src") {
            img = rdr.getValue();
        } else if (rdr.getName() == "align") {
            align = rdr.getValue();
        }
        next();
    }
    while (current != rdr.TagEnd && current != rdr.Eof) {
        next();
    }

    Ptr<GfxPixmap> pix = requestPixmap(ResId(img));
    if (pix) {
        std::auto_ptr<RichDocumentBlockObject> obj(new RichDocumentImageObject(pix));
        if (align == "left") {
            doc.addFloatObject(obj, true);
        } else if (align == "right") {
            doc.addFloatObject(obj, false);
        } else {
            doc.addCenterObject(obj);
        }
    }
}

/** Parse table. Upon entry, the "table" tag has been read. Upon exit,
    we're looking at the closing tag. */
void
RichHelpRenderer::parseTable()
{
    /* Parse table attributes */
    int align = 1;
    while (current == rdr.TagAttribute) {
        if (rdr.getName() == "align") {
            if (rdr.getValue() == "left") {
                align = 0;
            } else if (rdr.getValue() == "right") {
                align = 2;
            } else {
                align = 1;
            }
        }
        next();
    }

    /* Parse table content */
    std::vector<int> cellWidths;
    while (current != rdr.TagEnd && current != rdr.Eof && current != rdr.Error) {
        if (isOpen("tr")) {
            parseTableLine(cellWidths, align);
            next();
        } else {
            parseUnknown();
        }
    }
    doc.addNewline();
}

void
RichHelpRenderer::parseTableLine(std::vector<int>& cellWidths, const int align)
{
    /* Parse */
    ptr_vector<RichText> cellText;
    std::vector<int> cellAlign;
    while (current != rdr.TagEnd && current != rdr.Eof && current != rdr.Error) {
        if (isOpen("th")) {
            parseTableCell(cellWidths, cellText, cellAlign, 1);
            next();
        } else if (isOpen("td")) {
            parseTableCell(cellWidths, cellText, cellAlign, 0);
            next();
        } else if (isOpen("tn")) {
            parseTableCell(cellWidths, cellText, cellAlign, 2);
            next();
        } else {
            parseUnknown();
        }
    }

    /* Found anything? */
    if (cellText.empty()) {
        return;
    }

    /* Compute layout */
    const int PADDING = 5;
    int totalWidth = PADDING*(cellText.size() - 1);
    for (std::size_t i = 0; i < cellText.size(); ++i) {
        totalWidth += cellWidths[i];
    }

    int lm = doc.getLeftMargin();
    int rm = doc.getRightMargin();
    int wi = doc.getPageWidth();
    int room = wi - lm - rm;
    int x;
    if (room < totalWidth) {
        x = lm;
    } else {
        x = lm + align*(room - totalWidth)/2;
    }

    /* Render table */
    for (std::size_t i = 0; i < cellText.size(); ++i) {
        doc.setLeftMargin(x);
        doc.setRightMargin(std::max(0, wi - x - cellWidths[i]));
        if (cellAlign[i] == 0) {
            /* Left */
            doc.addAt(x, *cellText[i]);
            doc.addWordSeparator();
        } else if (cellAlign[i] == 1) {
            /* Center */
            doc.addCentered(x + cellWidths[i]/2, *cellText[i]);
        } else {
            /* Right */
            doc.addRight(x + cellWidths[i], *cellText[i]);
        }
        x += cellWidths[i];
        x += PADDING;
    }
    doc.setLeftMargin(lm);
    doc.setRightMargin(rm);
    doc.addNewline();
}

void
RichHelpRenderer::parseTableCell(std::vector<int>& cellWidths,
                                 ptr_vector<RichText>& cellText,
                                 std::vector<int>& cellAlign,
                                 const int align)
{
    /* Tag attributes */
    int userAlign = align;
    int userWidth = 0;
    while (current == rdr.TagAttribute) {
        if (rdr.getName() == "align") {
            if (rdr.getValue() == "left") {
                userAlign = 0;
            } else if (rdr.getValue() == "right") {
                userAlign = 2;
            } else {
                userAlign = 1;
            }
        }
        if (rdr.getName() == "width") {
            if (!stringValue(rdr.getValue(), userWidth)) {
                userWidth = 0;
            }
        }
        next();
    }

    /* Text content */
    RichText thisText(parseText(false));

    /* Set cell width */
    if (cellText.size() >= cellWidths.size()) {
        /* We do not yet have a size for this cell. If user gave one, convert
           to ems; otherwise estimate.*/
        if (userWidth != 0) {
            userWidth *= font_ems[FONT_NORMAL_BOLD];
        } else {
            userWidth = fonts[FONT_NORMAL_BOLD]->getTextWidth(thisText.getText());
        }
        cellWidths.push_back(userWidth);
    }

    /* Remember it. align=1 is <th>, which is implicitly bold. */
    if (align == 1) {
        cellText.push_back(new RichText(thisText.withAttribute(RichTextStyleAttribute::Bold)));
    } else {
        cellText.push_back(new RichText(thisText));
    }
    cellAlign.push_back(userAlign);
}

void
RichHelpRenderer::parsePre()
{
    /* Tag attributes */
    string_t kind;
    while (current == rdr.TagAttribute) {
        if (rdr.getName() == "class") {
            kind = rdr.getValue();
        }
        next();
    }

    /* Read it */
    RichText content(parseText(true));

    /* Parse it */
    if (kind == "ccscript") {
        RichText out;
        renderScript(content, out);
        out.swap(content);
    }

    doc.setLeftMargin(doc.getLeftMargin() + font_ems[FONT_NORMAL]*2);
    doc.addPreformatted(content.withAttribute(RichTextStyleAttribute::Fixed));
    doc.addNewline();
    doc.setLeftMargin(doc.getLeftMargin() - font_ems[FONT_NORMAL]*2);
}


void
RichHelpRenderer::parseDocument()
{
    /* A document is a list of block elements.
       It might be intermixed with
       - remaining attributes from the opening tag
       - spurious text elements (usually blank) */
    current = rdr.read();
    while (current != rdr.TagEnd && current != rdr.Eof && current != rdr.Error) {
        if (current == rdr.TagStart) {
            parseBlock(0);
        } else {
            next();
        }
    }
}

/**************************** RichTextParser ***************************/

inline
RichTextParser::RichTextParser(XmlReader& rdr)
    : RichParserBase(rdr)
{ }

RichText
RichTextParser::parse()
{
    // FIXME: like the <p> handling, this is almost a copy of parseText.
    bool haveSpace = true;
    RichText result;
    while (1) {
        if (current == rdr.Eof || current == rdr.Error || current == rdr.TagEnd) {
            /* EOF or error stop parsing, or surrounding tag ends */
            break;
        } else if (current == rdr.TagAttribute) {
            /* Spurious attribute from surrounding tag */
            next();
        } else if (current == rdr.Text) {
            /* Text */
            appendText(result, haveSpace, rdr.getValue());
            next();
        } else if (isOpen("br")) {
            /* Newline */
            result.append("\n\n");
            while (current != rdr.Eof && current != rdr.Error && current != rdr.TagEnd) {
                next();
            }
            next();
            haveSpace = true;
        } else {
            result.append(parseTextItem(false));
            next();
            haveSpace = false;
        }
    }
    return result;
}

/****************************** Entry Point ******************************/

/** Load help page into a rich document.
    \param doc [out] The document
    \param rdr [in] XML Reader
    \param name [in] Name of help page */
void
loadHelpPage(RichDocument& doc, XmlReader& rdr, const string_t name)
{
    /* Locate help page */
    bool atPageStart = false;
    enum { Searching, Found, End } state = Searching;
    while (state == Searching) {
        switch (rdr.read()) {
         case XmlReader::TagStart:
            atPageStart = (rdr.getTag() == "page");
            break;
         case XmlReader::TagAttribute:
            if (atPageStart && rdr.getName() == "id" && rdr.getValue() == name) {
                state = Found;
            }
            break;
         case XmlReader::Eof:
            state = End;
            break;
         default:
            break;
        }
    }

    /* Found? */
    if (state != Found) {
        /* Not found. Render error page. */
        doc.add(RichText(_("Error")).withAttribute(RichTextStyleAttribute::Big).withAttribute(RichTextStyleAttribute::Bold));
        doc.addParagraph();
        doc.add(format(_("The help page \"%s\" could not be found within the help file."), name));
        doc.finish();
    } else {
        /* Found. Read and render it. */
        RichHelpRenderer(doc, rdr).parseDocument();
        doc.finish();
    }
}

/** Parse XML into RichDocument. The text can contain all allowed markup. */
void
parseRichDocument(RichDocument& doc, string_t text)
{
    MemoryStream ms(text);
    XmlReader rdr(ms);
    RichHelpRenderer p(doc, rdr);
    p.next();
    p.parseDocument();
    doc.finish();
}

/** Parse XML into rich-text. The text can contain flow-text markup, i.e. \<b>, \<em>, etc. */
RichText
parseRichText(string_t text)
{
    MemoryStream ms(text);
    XmlReader rdr(ms);
    RichTextParser p(rdr);
    p.next();
    return p.parse();
}
